/***************************************************************************
                          disk.h  -  description
                             -------------------
    begin                :   5 6 2004
    copyright            : (C) 2004 by Naoya OYAMA
    email                : naoyan@minaminoshima.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <sys/types.h>

#define DISK_MAX              32
#define DELIMITER             " -- "
#define DEV_DIRECTORY         "/dev/"
#define DEVRANDOM             "/dev/urandom"
#define DEVZERO               "/dev/zero"
#define DEVNULL               "/dev/null"
#define PROC_MOUNTS           "/proc/mounts"
#define PROC_SWAPS            "/proc/swaps"
#define DEFAULT_SECTOR_SIZE   512
#define MAX_SECTOR_SIZE       2048
#define SECTOR_SIZE           512     /* still used in BSD code */
#define BUFFERS               32
#define BUFFER_ALLOC_SIZE     (1024*4*128)
#define BUFFER_SIZE           (1024*4*3*42)
#define WORK_UNIT_SIZE        (BUFFER_SIZE*16)
#define MAXIMUM_PARTS         60
#define NSA_COUNT             2
#define BOTTOM_MARGIN         60
#define TOP_MARGIN            310
#define LEFT_MARGIN           20
#define RIGHT_MARGIN          20
#define BADSECTOR_LOG         "/tmp/kdshred_badsector.log"
#define INFOREFRESH           5000 /* disk refresh timing (msec) */
#define PIXMAPREFRESH         50   /* pixmap refresh timing (msec) */
#define __INITIALIZE__        1    /* 0 : read(2)
                                      1 : write(2) */
#define UNABLE_TO_WRITESECTOR 0x00000001
#define UNABLE_TO_READSECTOR  0x00000002
#define VERIFY_ERROR          0x00000004

enum failure {usage, usage2, ioctl_error,
	unable_to_open, unable_to_read, unable_to_seek,
	unable_to_write, out_of_memory};
enum pattern {randomize, zero, one, two, three, four, five, six, seven,
	      eight, nine, aaaa, bbbb, cccc, dddd, eeee, ffff, gutmann00,
	      gutmann01, gutmann02, gutmann03, gutmann04, gutmann05, 
	      dbdb, dbdb_complement};
enum shredmethod {nsa, zero_clear, ff, random_only, random_zero, DoD5220_22_M,
		  AR380_19, AFSSI_5020, NAVSO_P_5239_26, NCSC_TG_025,
		  GUTMANN };

typedef struct {
	unsigned int heads;
	unsigned int sectors;
	unsigned int cylinders;
} geom;

typedef struct {
	unsigned char boot_ind;         /* 0x80 - active */
	unsigned char head;             /* starting head */
	unsigned char sector;           /* starting sector */
	unsigned char cyl;              /* starting cylinder */
	unsigned char sys_ind;          /* What partition type */
	unsigned char end_head;         /* end head */
	unsigned char end_sector;       /* end sector */
	unsigned char end_cyl;          /* end cylinder */
	unsigned char start4[4];        /* starting sector counting from 0 */
	unsigned char size4[4];         /* nr of sectors in partition */
} partition;

typedef struct {
	char file[128];
	char system[128];
	char fs[128];
	int partition_status; /* partition status info
				0  : good partition
				-1 : bad partition */
	unsigned int start_cyl;
	unsigned int end_cyl;
	unsigned int capa;              /* in MiB */
	int is_extended;
	int is_mounted;
} part_info;

typedef struct {
	partition *part_table;
	partition *ext_pointer;
	char changed;
	unsigned int offset;
	unsigned char *sectorbuffer;
} pte;

typedef struct {
	unsigned char type;
	char *name;
} systypes;

typedef struct {
	void *next;
	off_t sector;
	unsigned char *sector_dump;
	int status;
	int io_len;
} bad_sector;

typedef struct {
	char *bus;
	char *device;
	char *desc;
	bad_sector *bad_sectors;
	unsigned int partitions;
	unsigned int sector_size;
	unsigned int cylinder_size;
	off_t total_number_of_sectors;
	char file[128];
	geom geometry;
	pte ptes[MAXIMUM_PARTS];
	part_info part_infos[MAXIMUM_PARTS];
} harddisk;

