/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <ctype.h>
#include "disk.h"
#include "hwprobe.h"
#include "dd.h"
#include "full_read.h"

/* function proto type */
int check_badsector(int fd, off_t seek, harddisk *disk, unsigned char **buffers, int len, int mode);
void add_badsector_chain(harddisk *disk, unsigned char *buf, off_t cur_sect, int status, int io_len);
void free_badsectors( harddisk *disk );

int dd(harddisk *disk, unsigned char **buffers, unsigned int cylinder, 
		unsigned int cyl_len, int flag, int do_verify)
{
#if 0
argv0 : harddisk *disk;
argv1 : char **buffers;
argv2 : unsigned int cylinder;
argv3 : unsigned int write length.(last cylinder is not same as cylonder length.);
argv4 : last cylinder flag(last cylinder is not alignment by 4k block).
#endif
	int fd, random_index, len, omode, unit_offset = 0, loop_cnt, *indexes, i;
	off_t seek = (off_t)cylinder * WORK_UNIT_SIZE;

	if(seteuid(0)!=0) exit(-1);
#if __INITIALIZE__
	omode = O_WRONLY;
#else
	omode = O_RDONLY;
#endif
	if( !flag )
		omode |= O_DIRECT;
	else
		omode |= O_SYNC;
	if((fd = open(disk->file, omode)) < 0)
		fatal(unable_to_open);
	if ( seek )
		if ( lseek(fd, seek, SEEK_SET) == (off_t) -1 )
			fatal(unable_to_seek);
	loop_cnt = (cyl_len + BUFFER_SIZE - 1)/BUFFER_SIZE;
	if ( (indexes = malloc(sizeof(int)*loop_cnt) ) == NULL )
		fatal(out_of_memory);
	i = 0;
	while( cyl_len - unit_offset > 0 ) {
		if ( cyl_len - unit_offset < BUFFER_SIZE )
			len = cyl_len - unit_offset;
		else
			len = BUFFER_SIZE;
		random_index = (int)((BUFFERS*(1.0))*rand()/(RAND_MAX+1.0));
		*(indexes+i) = random_index;
#if __INITIALIZE__
		if ( write(fd, buffers[random_index], len) != len) {
#else
		if ( read(fd, buffers[random_index], len) != len) {
#endif
			close(fd);
			if((fd = open(disk->file, omode&~O_DIRECT)) < 0)
				fatal(unable_to_open);
			check_badsector(fd, seek+unit_offset, disk, buffers, len, 1);
			close(fd);
			if((fd = open(disk->file, omode)) < 0)
				fatal(unable_to_open);
		}
		unit_offset += len;
		i++;
	}
	close(fd);
	if(seteuid(getuid())!=0) exit(-1);
	if ( do_verify )
		verify_unit(disk, buffers, cylinder, cyl_len, flag, indexes);
	free(indexes);
	return 0;
}

int verify_unit(harddisk *disk, unsigned char **buffers, 
		unsigned int unit_number, unsigned int cyl_len, int flag, 
		int *indexes)
{
#if 0
argv0 : harddisk *disk;
argv1 : char **buffers;
argv2 : unsigned int unit number;
argv3 : unsigned int write length.(last cylinder is not same as cylonder length.);
argv4 : last cylinder flag(last cylinder is not alignment by 4k block).
#endif
	int fd, len, omode, unit_offset = 0, cnt, i;
	unsigned char *p0, *p1;
	off_t seek = (off_t)unit_number * WORK_UNIT_SIZE;

	if(seteuid(0)!=0) exit(-1);
	omode = O_RDONLY;
	if( !flag )
		omode |= O_DIRECT;
	else
		omode |= O_SYNC;
	if((fd = open(disk->file, omode)) < 0)
		fatal(unable_to_open);
	if ( seek )
		if ( lseek(fd, seek, SEEK_SET) == (off_t) -1 )
			fatal(unable_to_seek);
	i = 0;
	while( cyl_len - unit_offset > 0 ) {
		if ( cyl_len - unit_offset < BUFFER_SIZE )
			len = cyl_len - unit_offset;
		else
			len = BUFFER_SIZE;
		if ( (read(fd, buffers[BUFFERS], len) != len)  ){
			close(fd);
			if((fd = open(disk->file, omode&~O_DIRECT)) < 0)
				fatal(unable_to_open);
			check_badsector(fd, seek+unit_offset, disk, buffers, len, 0);
			close(fd);
			if((fd = open(disk->file, omode)) < 0)
				fatal(unable_to_open);
		}
		cnt = 0;
		p0 = buffers[BUFFERS];
		p1 = buffers[*(indexes+i)];
		while( cnt < len ) {
			if ( memcmp(p0, p1, disk->sector_size) )
				add_badsector_chain( disk, p0,
					(seek+unit_offset+cnt)/disk->sector_size,
					VERIFY_ERROR, disk->sector_size);
			p0  += disk->sector_size;
			p1  += disk->sector_size;
			cnt += disk->sector_size;
		}
		i++;
		unit_offset += len;
	}
	close(fd);
	if(seteuid(getuid())!=0) exit(-1);
	return 0;
}

int creat_buffers(unsigned char **buffers, enum pattern p)
{
	int i, fd, psize;
	unsigned char gutmann_buf[3];

#if __BSD_SOURCE
	psize = getpagesize();
#else
	psize = sysconf(_SC_PAGE_SIZE);
#endif
	for ( i=0; i < BUFFERS+1; i++ ) {
		if ( posix_memalign(&buffers[i], psize, BUFFER_ALLOC_SIZE) != 0 )
			fatal(out_of_memory);
	}
	switch ( p ){
		case randomize:
			if ((fd = open(DEVRANDOM, O_RDONLY)) < 0 )
				fatal(unable_to_open);
			for ( i=0; i < BUFFERS; i++ ){
#if __INITIALIZE__
				if ((full_read(fd, buffers[i], BUFFER_SIZE)) != BUFFER_SIZE)
					fatal(unable_to_read);
#endif
			}
			close(fd);
			break;

		case zero:
			set_buffers(buffers, 0);
			break;

		case one:
			set_buffers(buffers, 0x11111111);
			break;

		case two:
			set_buffers(buffers, 0x22222222);
			break;

		case three:
			set_buffers(buffers, 0x33333333);
			break;

		case four:
			set_buffers(buffers, 0x44444444);
			break;

		case five:
			set_buffers(buffers, 0x55555555);
			break;

		case six:
			set_buffers(buffers, 0x66666666);
			break;

		case seven:
			set_buffers(buffers, 0x77777777);
			break;

		case eight:
			set_buffers(buffers, 0x88888888);
			break;

		case nine:
			set_buffers(buffers, 0x99999999);
			break;

		case aaaa:
			set_buffers(buffers, 0xaaaaaaaa);
			break;

		case bbbb:
			set_buffers(buffers, 0xbbbbbbbb);
			break;

		case cccc:
			set_buffers(buffers, 0xcccccccc);
			break;

		case dddd:
			set_buffers(buffers, 0xdddddddd);
			break;

		case eeee:
			set_buffers(buffers, 0xeeeeeeee);
			break;

		case ffff:
			set_buffers(buffers, 0xffffffff);
			break;

		case dbdb:
			set_buffers(buffers, 0xdbdbdbdb);
			break;

		case dbdb_complement:
			set_buffers(buffers, ~0xdbdbdbdb);
			break;

		case gutmann00:
			gutmann_buf[0] = 0x92;
			gutmann_buf[1] = 0x49;
			gutmann_buf[2] = 0x24;
			set_buffers_24(buffers, gutmann_buf);
			break;

		case gutmann01:
			gutmann_buf[0] = 0x49;
			gutmann_buf[1] = 0x24;
			gutmann_buf[2] = 0x92;
			set_buffers_24(buffers, gutmann_buf);
			break;

		case gutmann02:
			gutmann_buf[0] = 0x24;
			gutmann_buf[1] = 0x92;
			gutmann_buf[2] = 0x49;
			set_buffers_24(buffers, gutmann_buf);
			break;

		case gutmann03:
			gutmann_buf[0] = 0x6d;
			gutmann_buf[1] = 0xb6;
			gutmann_buf[2] = 0xdb;
			set_buffers_24(buffers, gutmann_buf);
			break;

		case gutmann04:
			gutmann_buf[0] = 0xb6;
			gutmann_buf[1] = 0xdb;
			gutmann_buf[2] = 0x6d;
			set_buffers_24(buffers, gutmann_buf);
			break;

		case gutmann05:
			gutmann_buf[0] = 0xdb;
			gutmann_buf[1] = 0x6d;
			gutmann_buf[2] = 0xb6;
			set_buffers_24(buffers, gutmann_buf);
			break;

		default:
			break;
	}      
	return 0;
}

int remove_buffers(unsigned char **buffers)
{
	int i;
	for ( i=0; i < BUFFERS+1; i++ )
		free(buffers[i]);
	return 0;
}

int _is_mounted(char *file)
{
#if 0
  grep -e file /proc/mounts /proc/swaps
#endif
  
	char cmdline[1024];
	int  ret_val = 0, i;
	char tmp[128];
	FILE *mounts, *swaps;

	if ((mounts = fopen(PROC_MOUNTS, "r")) == NULL )
		fatal(unable_to_open);
	while ( fgets(cmdline, sizeof(cmdline), mounts) != NULL ) {
		i = 0;
		while ( i < sizeof(tmp)-2 ){
			if ( *(cmdline+i) == ' ' )
				break;
			tmp[i] = *(cmdline+i);
			i++;
		}
		tmp[i] = '\0';
		if ( strcmp(tmp, file) == 0 )
			ret_val = 1;
	}
	fclose(mounts);

	if ((swaps = fopen(PROC_SWAPS, "r")) == NULL )
		fatal(unable_to_open);
	while ( fgets(cmdline, sizeof(cmdline), swaps) != NULL ) {
		i = 0;
		while ( i < sizeof(tmp)-2 ){
			if ( *(cmdline+i) == ' ' )
				break;
			tmp[i] = *(cmdline+i);
			i++;
		}
		tmp[i] = '\0';
		if ( strcmp(tmp, file) == 0 )
			ret_val = 1;
	}
	fclose(swaps);

	return (ret_val);
}

int is_mounted(harddisk *disk)
{
#if 0
  grep -e disk->file /proc/mounts /proc/swaps
#endif
	int  i;

	i = 0;
	while ( i < disk->partitions ) {
		if( disk->part_infos[i].is_mounted )
			return(1);
		i++;
	}
	return (0);
}

int check_badsector(int fd, off_t seek, harddisk *disk, unsigned char **buffers, int len, int mode)
{
#if 0
mode : 1 write
       0 read
#endif
	int write_size = 0, random_index, io_len;
	off_t cur_sect;
	bad_sector * bad_sect;
	char buf[MAX_SECTOR_SIZE];
	int status;

	if ( lseek(fd, seek, SEEK_SET) == (off_t) -1 )
		fatal(unable_to_seek);
	while( write_size < len ) {
		random_index = (int)((BUFFERS*(1.0))*rand()/(RAND_MAX+1.0));
		if ( mode == 1 )
#if __INITIALIZE__
			io_len =  write(fd, buffers[random_index],
						disk->sector_size);
#else
			io_len = read(fd, buffers[random_index],
						disk->sector_size);
#endif
		else
			io_len = read(fd, buffers[random_index],
						disk->sector_size);
		if ( io_len != disk->sector_size) {
			if ( mode == 1 )
				status = UNABLE_TO_WRITESECTOR;
			else
				status = UNABLE_TO_READSECTOR;
			bad_sect = disk->bad_sectors;
			cur_sect = (seek+write_size)/disk->sector_size;
			if ( lseek(fd, cur_sect*disk->sector_size, SEEK_SET) == (off_t) -1 )
				fatal(unable_to_seek);
			if ( io_len = read(fd, buf, disk->sector_size) != 
							disk->sector_size)
				status |= UNABLE_TO_READSECTOR;
			add_badsector_chain(disk, buf, cur_sect, status, io_len);
		}
		write_size += disk->sector_size;
	}
	return 0;
}

void add_badsector_chain(harddisk *disk, unsigned char *buf,
		off_t cur_sect, int status, int io_len)
{
	bad_sector *bad_sect;
	if ( disk->bad_sectors == NULL ){
		if( (disk->bad_sectors = malloc(sizeof(bad_sector))) == NULL )
			fatal(out_of_memory);
		bad_sect = disk->bad_sectors;
		memset(bad_sect, 0, sizeof(bad_sect));
		bad_sect->sector = -1;
	}
	else
		bad_sect = disk->bad_sectors;

	while ( bad_sect->sector != -1 ){
		if ( bad_sect->sector == cur_sect ){
			bad_sect->status |= status;
			memset(bad_sect->sector_dump, 0, disk->sector_size);
			memcpy(bad_sect->sector_dump, buf, io_len);
			bad_sect->io_len = io_len;
			return;
		}
		bad_sect = (bad_sector *)bad_sect->next;
	}
	if((bad_sect->sector_dump = malloc(disk->sector_size)) == NULL )
		fatal(out_of_memory);
	memset(bad_sect->sector_dump, 0, disk->sector_size);
	memcpy(bad_sect->sector_dump, buf, io_len);
	bad_sect->sector = cur_sect;
	bad_sect->status = status;
	bad_sect->io_len = io_len;
	if ( (bad_sect->next = malloc(sizeof(bad_sector))) == NULL )
		fatal(out_of_memory);
	memset(bad_sect->next, 0, sizeof(bad_sector));
	((bad_sector*)bad_sect->next)->sector = -1;

	return;
}

int write_badsector_log( harddisk *disk ){
	FILE *fp;
	bad_sector *bad_sect;
	unsigned char *p, tmp[17], c;
	int i;
	off_t offset;
	
	if(seteuid(0)!=0) exit(-1);
	if ( (fp = fopen(BADSECTOR_LOG, "w")) == NULL ) {
		return -1;
	}
	bad_sect = disk->bad_sectors;
	while ( bad_sect->sector != -1 ){
		fprintf(fp, "sector #%lld ", bad_sect->sector);
		if ( bad_sect->status & UNABLE_TO_WRITESECTOR )
			fprintf(fp, "/unable to write sector");
		if ( bad_sect->status & UNABLE_TO_READSECTOR )
			fprintf(fp, "/unable to read sector");
		if ( bad_sect->status & VERIFY_ERROR )
			fprintf(fp, "/verify error");
		fprintf(fp, "/IO len = %d\n", bad_sect->io_len);
		offset = bad_sect->sector * disk->sector_size;
		i = 0;
		p = bad_sect->sector_dump;
		while( i < bad_sect->io_len ) {
			if ( (i%16) == 0 )
				fprintf(fp, "%012llx ", offset);
			c = *(p+i);
			fprintf(fp, "%02x ", c);
			if ( isprint(*(p+i)) )
				tmp[i%16] = c;
			else
				tmp[i%16] = '.';
			if( (i%16) == 15 ){
				tmp[sizeof(tmp)-1] = '\0';
				fprintf(fp, "  %s\n", tmp);
			}
			i++;
			offset++;
		}
		fputc('\n', fp);
		bad_sect = (bad_sector*)bad_sect->next;
	}
	fclose(fp);
	if(seteuid(getuid())!=0) exit(-1);
	free_badsectors( disk );
	return 0;
}

int sort_part_info( const void *pi1, const void *pi2 )
{
	part_info *part_info1 = (part_info*)pi1,
		*part_info2 = (part_info*)pi2;
	return ( part_info1->start_cyl - part_info2->start_cyl );
}

void set_buffers( unsigned char **buffers, const int c )
{
	int i;
	for( i = 0; i < BUFFERS; i++ )
		memset(buffers[i], c, BUFFER_SIZE);
}

void set_buffers_24( unsigned char **buffers, const unsigned char c[3] )
{
	int i, j;
	for ( i = 0; i < BUFFERS; i++ )
		for( j = 0; j < BUFFER_SIZE; j++ )
			buffers[i][j] = c[j%3];
}


void free_badsectors( harddisk *disk )
{
	bad_sector *p0, *p1;

	p0 = (bad_sector *)disk->bad_sectors;
	while( p0->sector != -1 ){
		p1 = (bad_sector *)p0->next;
		free(p0->sector_dump);
		free(p0);
		p0 = p1;
	}
	free(p0);
}
 

int dump_sector( harddisk *disk, off_t sector )
{
	int fd, i;
	char buf[MAX_SECTOR_SIZE];
	char tmp[17];
	unsigned char *p, c;
	off_t offset = sector * disk->sector_size;

	if((fd = open(disk->file, O_RDONLY)) < 0)
		fatal(unable_to_open);
	if ( lseek(fd, offset, SEEK_SET) == (off_t) -1 )
		fatal(unable_to_seek);
	read(fd, buf, disk->sector_size);

	i = 0;
	p = buf;
	while( i < disk->sector_size ) {
		if ( (i%16) == 0 )
			printf("%012llx ", offset);
		c = *(p+i);
		printf("%02x ", c);
		if ( isprint(*(p+i)) )
			tmp[i%16] = c;
		else
			tmp[i%16] = '.';
		if( (i%16) == 15 ){
			tmp[sizeof(tmp)-1] = '\0';
			printf("  %s\n", tmp);
		}
		i++;
		offset++;
	}
	return(0);
}
